#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "defines.h"
#include "trd.h"
#include "exception_t.h"

#define ARGS_MAX 10
#define VERSION "0.1b1"

static void usage()
{
	printf("trdetz the .trd file rummager\nver %s by boo_boo [boo_boo(^at^)inbox.ru]\n",VERSION);
	printf("\n");
	printf("USAGE: trdetz [flags] command [arguments]\n");
	printf("Numeric args may be given as [dec], 0x[hex], or 0[oct]\n");
	printf("\n");
	printf("Flags:\n");
	printf("-df[=del_file_marker] : handle deleted files (and optionally set deleted file marker, which is '?' by default)\n");
	printf("-fs : use num_sectors*256 as file length\n");	
	printf("\n");
	printf("Commands:\n");
	printf("dinfo <trd> : print disc summary\n");
	printf("cat <trd>, [fname] : print catalogue (or single file) info\n");
	printf("secread <trd> <bin_fname> <track> <sector> <num_sectors> : copy sectors from trd to binary file\n");
	printf("secwrite <trd> <bin_fname> <track> <sector> <num_sectors> : copy sectors from binary file to trd\n");
	printf("fread <trd> <fname> [bin_fname] : copy file from trd to binary\n");
	printf("fwrite <trd> <bin_fname> [fname] : copy binary file to trd\n");
	printf("del <trd> <fname> : delete file from trd catalogue\n");
	printf("ren <trd> <old_fname> <new_fname> : rename file in trd catalogue (use with -df to undelete file)\n");
	printf("fprop <trd> <fname> <property name> <property value> : change file property\n");
	printf("       where property name is one of:\n");
	printf("         start -- for file start adress\n");
	printf("         flen -- for file length in bytes\n");
	printf("         nsec -- for file length in sectors\n");
	printf("\n");
	
	return;
}

static char *cmd_args[ARGS_MAX];
static int cmd_args_num=0;

static void cmd_getargs(int argc, char *argv[], int from)
{
	int i,n;
	
	cmd_args_num=0;
	for(n=0; n < ARGS_MAX; ++n) cmd_args[n] = NULL;
	
	for(i=from, n=0; i < argc; ++i, ++n)
	{
		cmd_args[n]=argv[i];
		cmd_args_num++;
	}
	
	return;
}

int main(int argc, char *argv[])
{
	char *command = NULL;
	char *endptr;
	char *fname;
	int clarg, i, prn, sector, track, num_sectors, num_bytes, ivalue, num_files;
	char strbuff[100];
	char with_del_files = 0;
	struct trd_cat_entry *pcat = NULL;
	char del_file_mark='?';
	char fname_ext_separator='.';
	unsigned use_cat_bytelen=1;
	struct trd_fname trd_filename, trd_filename2;
	BYTE *bfile_buff = NULL;
	BYTE trd_buff[TRD_SIZE];
	FILE *bfp = NULL;
	int is_last_file;
	struct exception e;
	
	init_exception_context(the_exception_context);
	
	if(argc < 2) goto usage;

	/*parse flags*/
	for(clarg=1; clarg < argc; ++clarg)
	{
		if(*(argv[clarg]) != '-') break;
		
		if(!strncmp(argv[clarg],"-df=",4) && strlen(argv[clarg])==5)
		{
			with_del_files=(argv[clarg])[4]; /*handle deleted files (as "[sym]rest_of_name")*/
		}
		else if(!strcmp(argv[clarg],"-df")) with_del_files='?'; /*handle deleted files (as "?rest_of_name")*/
		else if(!strcmp(argv[clarg],"-fs")) use_cat_bytelen=0;	/*use file length as given in the catalogue instead of num_sectors*256*/
		else 
		{
			fprintf(stderr, "Unknown flag: %s\n", argv[clarg]);
			goto fail;
		}
	}
	
	if(clarg < argc) command=argv[clarg];
	else goto usage;
	cmd_getargs(argc, argv, ++clarg);
	
	if(!strcmp(command,"dinfo")) /*dinfo <trd> : print disc summary*/
	{
		Try trd_read(cmd_args[0], trd_buff);
		Catch(e) goto exc_fail;
		
		printf("title: %s\n",(trd_buff+(0x100*8)+0xF5));
		printf("free_sectors: %d\n",TO_WORD(*(trd_buff+(0x100*8)+0xE6),*(trd_buff+(0x100*8)+0xE5)));
		printf("files: %d\n",*(trd_buff+(0x100*8)+0xE4));
		printf("deleted_files: %d\n",*(trd_buff+(0x100*8)+0xF4));
		printf("next_free_track: %d\n",*(trd_buff+(0x100*8)+0xE2));
		printf("next_free_sector: %d\n",*(trd_buff+(0x100*8)+0xE1));
		
	}
	else if(!strcmp(command,"cat")) /*cat <trd>, [fname] : print catalogue (or single file) info*/
	{
		if(cmd_args_num < 1) goto cmd_args_fail;
		
		Try
		{
			trd_read(cmd_args[0], trd_buff);
			
			if(cmd_args[1] != NULL) trd_parse_fname(cmd_args[1], fname_ext_separator, with_del_files, &trd_filename);
		
			prn=0;
			num_files=*(trd_buff+(0x100*8)+0xE4);
			for(i=0, pcat=(struct trd_cat_entry *)trd_buff; i< num_files; ++i, ++pcat)
			{
				if(!((pcat->filename.name)[0]) || ((pcat->filename.name)[0]) == 0x01)
				{
					if(!with_del_files) continue;
					
					strncpy(strbuff+1, (char *)(pcat->filename.name + 1), 7);
					*strbuff = del_file_mark;
				}
				else strncpy(strbuff, (char *)pcat->filename.name, 8);
				strbuff[8]='\0';	
				
				if(cmd_args[1] != NULL)
				{
					if(!trd_fname_equal(&trd_filename, &(pcat->filename))) continue;
				}
					
				printf("%s.%c %3u %05d %05d %3u %3u\n",
					strbuff, 
					(char)(pcat->filename.ext), 
					pcat->sectors, 
					TO_WORD(pcat->start_h, pcat->start_l), 
					TO_WORD(pcat->length_h, pcat->length_l),
					pcat->start_track,
					pcat->start_sector
					);
				prn=1;
			}	
		}
		Catch(e) goto exc_fail;

		if(cmd_args[1] != NULL && !prn)
		{
			fprintf(stderr, "File '%s' not found\n", cmd_args[1]);
			goto fail;			
		}
	}
	else if(!strcmp(command, "secwrite")) /*secwrite <trd> <bin_fname> <track> <sector> <num_sectors> : copy sectors from binary file to trd*/
	{
		if(cmd_args_num < 5) goto cmd_args_fail;
		
		track = strtol(cmd_args[2],&endptr,0);
		if(*endptr != '\0' || track < 0) goto cmd_args_fail;

		sector = strtol(cmd_args[3],&endptr,0);
		if(*endptr != '\0' || sector < 0) goto cmd_args_fail;
			
		num_sectors = strtol(cmd_args[4],&endptr,0);
		if(*endptr != '\0' || num_sectors < 1) goto cmd_args_fail;
		
		/*if(num_sectors > 0x100)
		{
			fprintf(stderr, "Too many sectors\n");
			goto fail;
		}*/			
		
		if(NULL == (bfile_buff = (BYTE *) malloc(num_sectors * 0x100)))
		{
			fprintf(stderr, "Failed to allocate memory\n");
			goto fail;			
		}
		memset(bfile_buff,0,num_sectors * 0x100);		

		if(NULL == (bfp = fopen(cmd_args[1],"rb")))
		{
			fprintf(stderr, "Cannot open file %s for read\n",cmd_args[1]);
			goto fail;			
		}

		if(!fread(bfile_buff,1,num_sectors * 0x100,bfp))
		{
			fprintf(stderr, "Cannot read from file %s\n",cmd_args[1]);
			goto fail;			
		}			
		
		fclose(bfp);
		
		Try
		{
			trd_read(cmd_args[0], trd_buff);
			trd_write_sectors(trd_buff, track, sector, num_sectors, bfile_buff);
			free(bfile_buff);
			trd_write(cmd_args[0], trd_buff);
		}
		Catch(e) goto exc_fail;
	}
	else if(!strcmp(command, "secread")) /*secread <trd> <bin_fname> <track> <sector> <num_sectors> : copy sectors from trd to binary file*/
	{
		if(cmd_args_num < 5) goto cmd_args_fail;

		track = strtol(cmd_args[2],&endptr,0);
		if(*endptr != '\0' || track < 0) goto cmd_args_fail;

		sector = strtol(cmd_args[3],&endptr,0);
		if(*endptr != '\0' || sector < 0) goto cmd_args_fail;
			
		num_sectors = strtol(cmd_args[4],&endptr,0);
		if(*endptr != '\0' || num_sectors < 1) goto cmd_args_fail;		
		
		if(NULL == (bfile_buff = (BYTE *) malloc(num_sectors * 0x100)))
		{
			fprintf(stderr, "Failed to allocate memory\n");
			goto fail;			
		}
		memset(bfile_buff,0,num_sectors * 0x100);
		
		Try
		{
			trd_read(cmd_args[0], trd_buff);
			trd_read_sectors(trd_buff, track, sector, num_sectors, bfile_buff);
		}
		Catch(e) goto exc_fail;
		
		if(NULL == (bfp = fopen(cmd_args[1],"wb")))
		{
			fprintf(stderr, "Cannot open file %s for writing\n",cmd_args[1]);
			goto fail;			
		}

		if(fwrite(bfile_buff,1,num_sectors * 0x100,bfp) != num_sectors * 0x100)
		{
			fprintf(stderr, "Error writing to file %s\n",cmd_args[1]);
			goto fail;			
		}		

		free(bfile_buff);
		fclose(bfp);
	}	
	else if(!strcmp(command, "fwrite")) /*fwrite <trd> <bin_fname> [fname]: copy binary file to trd*/
	{
		if(cmd_args_num < 2) goto cmd_args_fail;
		
		if(NULL == (bfp = fopen(cmd_args[1],"rb")))
		{
			fprintf(stderr, "Cannot open file %s for reading\n",cmd_args[1]);
			goto fail;			
		}

		if(fseek(bfp, 0, SEEK_END))
		{
			fprintf(stderr, "Cannot seek to the end of file %s\n",cmd_args[1]);
			goto fail;			
		}
		
		num_bytes = ftell(bfp);
		rewind(bfp);
		
		if(NULL == (bfile_buff = (BYTE *) malloc(num_bytes)))
		{
			fprintf(stderr, "Failed to allocate memory\n");
			goto fail;			
		}
		
		if(!fread(bfile_buff,1,num_bytes,bfp))
		{
			fprintf(stderr, "Cannot read from file %s\n",cmd_args[1]);
			goto fail;			
		}

		if(cmd_args_num > 2) fname = cmd_args[2];
		else fname = cmd_args[1];

		Try
		{
			trd_parse_fname(fname,fname_ext_separator, with_del_files, &trd_filename);
			trd_read(cmd_args[0], trd_buff);
			trd_writefile(trd_buff, bfile_buff, num_bytes, &trd_filename);
			trd_write(cmd_args[0], trd_buff);
		}
		Catch(e) goto exc_fail;
		
		free(bfile_buff);
		fclose(bfp);			
	}
	else if(!strcmp(command, "fread")) /*fread <trd> <fname> [bin_fname] : copy file from trd to binary*/
	{
		if(cmd_args_num < 2) goto cmd_args_fail;
		
		Try
		{
			trd_read(cmd_args[0], trd_buff);
			trd_parse_fname(cmd_args[1],fname_ext_separator, with_del_files, &trd_filename);
		}
		Catch(e) goto exc_fail;
		
		if(NULL == (pcat = trd_find_cat_entry(trd_buff, &trd_filename, with_del_files, &is_last_file)))
		{
			fprintf(stderr, "File '%s' not found\n", cmd_args[1]);
			goto fail;			
		}
		
		if(use_cat_bytelen)
		{
			num_bytes = TO_WORD(pcat->length_h, pcat->length_l);
			num_sectors = ceil((float)num_bytes / 0x100);
		}
		else num_bytes = (num_sectors = pcat->sectors) * 0x100;
	
		if(NULL == (bfile_buff = (BYTE *) malloc(num_sectors * 0x100)))
		{
			fprintf(stderr, "Failed to allocate memory\n");
			goto fail;			
		}
		memset(bfile_buff,0,num_bytes);
		
		Try trd_read_sectors(trd_buff, pcat->start_track, pcat->start_sector, num_sectors, bfile_buff);
		Catch(e) goto exc_fail;

		if(cmd_args_num > 2) fname = cmd_args[2];
		else fname = cmd_args[1];
		
		if(NULL == (bfp = fopen(fname,"wb")))
		{
			fprintf(stderr, "Cannot open file %s for writing\n",fname);
			goto fail;			
		}

		if(fwrite(bfile_buff,1,num_bytes,bfp) != num_bytes)
		{
			fprintf(stderr, "Error writing to file %s\n",cmd_args[1]);
			goto fail;			
		}		

		free(bfile_buff);
		fclose(bfp);
	}
	else if(!strcmp(command, "del")) /*del <trd> <fname> : delete file from trd catalogue*/
	{
		if(cmd_args_num < 2) goto cmd_args_fail;
		
		Try
		{
			trd_read(cmd_args[0], trd_buff);
			trd_parse_fname(cmd_args[1],fname_ext_separator, with_del_files, &trd_filename);
		}
		Catch(e) goto exc_fail;
		
		if(NULL == (pcat = trd_find_cat_entry(trd_buff, &trd_filename, with_del_files, &is_last_file)))
		{
			fprintf(stderr, "File '%s' not found\n", cmd_args[1]);
			goto fail;			
		}
		
		Try
		{
			trd_deletefile(trd_buff, pcat, is_last_file);
			trd_write(cmd_args[0], trd_buff);			
		}
		Catch(e) goto exc_fail;
	}
	else if(!strcmp(command, "ren")) /*ren <trd> <old_fname> <new_fname> : rename file in trd catalogue*/
	{
		if(cmd_args_num < 3) goto cmd_args_fail;
		
		Try
		{
			trd_read(cmd_args[0], trd_buff);
			trd_parse_fname(cmd_args[1],fname_ext_separator, with_del_files, &trd_filename);
			trd_parse_fname(cmd_args[2],fname_ext_separator, with_del_files, &trd_filename2);
		}
		Catch(e) goto exc_fail;
		
		if(NULL == (pcat = trd_find_cat_entry(trd_buff, &trd_filename, with_del_files, &is_last_file)))
		{
			fprintf(stderr, "File '%s' not found\n", cmd_args[1]);
			goto fail;			
		}
		
		memcpy(&(pcat->filename),&trd_filename2,sizeof(struct trd_fname));
		
		Try trd_write(cmd_args[0], trd_buff);
		Catch(e) goto exc_fail;
	}	
	else if(!strcmp(command, "fprop")) /*fprop <trd> <fname> <property_name> <property value>: change file property*/
	{
		if(cmd_args_num < 4) goto cmd_args_fail;
		
		Try
		{
			trd_read(cmd_args[0], trd_buff);
			trd_parse_fname(cmd_args[1],fname_ext_separator, with_del_files, &trd_filename);
		}
		Catch(e) goto exc_fail;
		
		if(NULL == (pcat = trd_find_cat_entry(trd_buff, &trd_filename, with_del_files, &is_last_file)))
		{
			fprintf(stderr, "File '%s' not found\n", cmd_args[1]);
			goto fail;			
		}
		
		if(!strcmp(cmd_args[2],"start")) /*start adress*/
		{
			ivalue = strtol(cmd_args[3],&endptr,0);
			if(*endptr != '\0' || ivalue < 0 || ivalue > 0xFFFF) goto cmd_args_fail;
			
			pcat->start_l = (ivalue & 0xFF);
			pcat->start_h = (ivalue >> 8);
		}
		else if(!strcmp(cmd_args[2],"flen")) /*file length*/
		{
			ivalue = strtol(cmd_args[3],&endptr,0);
			if(*endptr != '\0' || ivalue < 0 || ivalue > 0xFFFF) goto cmd_args_fail;
			
			pcat->length_l = (ivalue & 0xFF);
			pcat->length_h = (ivalue >> 8);
		}
		else if(!strcmp(cmd_args[2],"nsec")) /*num of sectors*/
		{
			ivalue = strtol(cmd_args[3],&endptr,0);
			if(*endptr != '\0' || ivalue < 0 || ivalue > 0xFF) goto cmd_args_fail;
			
			pcat->sectors = ivalue;
		}		
		else goto cmd_args_fail;
		
		Try trd_write(cmd_args[0], trd_buff);
		Catch(e) goto exc_fail;
	}	
	else
	{	
		fprintf(stderr, "Unknown command: %s\n", command);
		goto fail;
	}
	
	return(EXIT_SUCCESS);

exc_fail:
	fprintf(stderr, "%s\n", e.msg);
fail:
	if(bfile_buff != NULL) free(bfile_buff);
	if(bfp != NULL) fclose(bfp);
	
	return(1);
	
cmd_args_fail:

	fprintf(stderr, "Invalid arguments for command '%s'\n", command);
	return(1);
	
usage:
	usage();
	return(EXIT_SUCCESS);
}
